/**
 * @file main.c
 * @brief Program to write circular buffer data to a valid HDF5 v0 file using Intel QAT and DSA.
 * @author Generated by Grok (xAI), checked by ChatGPT4o (OpenAI)
 * @date 2025-03-31
 *
 * Creates an HDF5 v0 file with a root group, B-tree, Name Heap, and compressed dataset.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <qatzip.h>
#include <dml/dml.h>

#define BUFFER_SIZE (1ULL << 30) ///< 1 billion floats (~8 GB)
#define CHUNK_SIZE (1 << 20)     ///< 1 MB chunk size
#define ALIGNMENT 64             ///< Alignment for DSA/QAT
#define MAX_COMPRESSED_SIZE(total_raw_size) ((total_raw_size) + 1024) ///< Margin for DEFLATE

dml_job_t* init_dsa() {
    dml_job_t *job;
    if (dml_init_job(DML_PATH_HW, &job) != DML_STATUS_OK) {
        fprintf(stderr, "Failed to initialize DSA\n");
        return NULL;
    }
    return job;
}

int dsa_write(dml_job_t *job, void *src, int fd, off_t offset, size_t size) {
    if (lseek(fd, offset + size - 1, SEEK_SET) == -1 || write(fd, "", 1) != 1) {
        perror("Error initializing file space");
        return -1;
    }
    lseek(fd, 0, SEEK_SET);

    job->operation = DML_OP_MEM_MOVE;
    job->source_first_ptr = src;
    job->destination_first_ptr = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, offset);
    if (job->destination_first_ptr == MAP_FAILED) {
        perror("mmap failed");
        return -1;
    }
    job->source_length = size;
    job->destination_length = size;

    dml_status_t status = dml_execute_job(job, DML_BLOCK_ON_FAULT);
    if (status != DML_STATUS_OK) {
        fprintf(stderr, "dml_execute_job failed: %d\n", status);
        munmap(job->destination_first_ptr, size);
        return -1;
    }

    munmap(job->destination_first_ptr, size);
    return 0;
}

void cleanup_dsa(dml_job_t *job) {
    dml_finalize_job(job);
}

uint64_t get_timestamp_ms() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (uint64_t)(tv.tv_sec) * 1000 + (uint64_t)(tv.tv_usec) / 1000;
}

void write_with_qat_dsa(float *circular_buffer, int head, int size, int capacity, const char *filename,
                        const char *label, uint32_t datatype, uint64_t dim_n, uint64_t dim_m,
                        double sampling_freq) {
    int fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, 0666);
    if (fd < 0) {
        perror("Error opening file");
        return;
    }

    // Calculate segments
    int start1 = (head - size >= 0) ? head - size : capacity - (size - head);
    int end1 = (head - size >= 0) ? head : capacity;
    int start2 = 0;
    int end2 = (head - size >= 0) ? 0 : head - size + capacity;

    size_t size1_raw = (end1 - start1) * sizeof(float);
    size_t size2_raw = (end2 - start2) * sizeof(float);
    size_t total_raw_size = size1_raw + size2_raw;

    if (total_raw_size > UINT32_MAX) {
        fprintf(stderr, "Error: Total size exceeds UINT32_MAX\n");
        close(fd);
        return;
    }

    uint32_t size1 = (uint32_t)size1_raw;
    uint32_t size2 = (uint32_t)size2_raw;

    // Initialize QAT
    QzSession_T session = {0};
    if (qzInit(&session, 1) != QZ_OK || qzSetupSession(&session, NULL) != QZ_OK) {
        fprintf(stderr, "Error initializing QAT\n");
        close(fd);
        return;
    }

    // Initialize DSA
    dml_job_t *dsa_job = init_dsa();
    if (!dsa_job) {
        qzClose(&session);
        close(fd);
        return;
    }

    // Compress data
    size_t max_compressed_size = MAX_COMPRESSED_SIZE(total_raw_size);
    unsigned char *compressed_data = aligned_alloc(ALIGNMENT, max_compressed_size);
    if (!compressed_data) {
        perror("Error in aligned_alloc");
        cleanup_dsa(dsa_job);
        qzClose(&session);
        close(fd);
        return;
    }
    size_t compressed_offset = 0;

    uint32_t comp_size1 = (uint32_t)(max_compressed_size - compressed_offset);
    if (qzCompress(&session, (unsigned char *)(circular_buffer + start1), &size1, 
                   compressed_data + compressed_offset, &comp_size1, 1) == QZ_OK) {
        compressed_offset += comp_size1;
    }

    uint32_t comp_size2 = (uint32_t)(max_compressed_size - compressed_offset);
    if (qzCompress(&session, (unsigned char *)(circular_buffer + start2), &size2, 
                   compressed_data + compressed_offset, &comp_size2, 1) == QZ_OK) {
        compressed_offset += comp_size2;
    }

    // HDF5 v0 Structure
    // Superblock (96 bytes)
    uint8_t superblock[96] = {
        0x89, 'H', 'D', 'F', 0x0D, 0x0A, 0x1A, 0x0A, // Signature
        0,    // Superblock version
        0,    // Free space version
        0,    // Root group symbol table version
        0,    // Reserved
        8, 0, 0, 0, // Size of Offsets
        8, 0, 0, 0, // Size of Lengths
        0, 0, 0, 0, // Reserved
        0, 0, 0, 0, // File Consistency Flags
        0, 0, 0, 0, 0, 0, 0, 0, // Base Address
        0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // Free Space Address
        0, 0, 0, 0, 0, 0, 0, 0, // End of File Address (updated later)
        0, 0, 0, 0, 0, 0, 0, 0  // Driver Info Block Address
        // Symbol Table Entry (36 bytes) follows
    };

    // Offsets
    uint64_t btree_offset = 96; // After superblock
    uint64_t name_heap_offset = btree_offset + 48; // After B-tree
    uint64_t dataset_offset = name_heap_offset + 64; // After Name Heap
    uint64_t data_offset = dataset_offset + 88; // After dataset header

    // Superblock Symbol Table Entry
    *(uint64_t *)(superblock + 60) = dataset_offset; // Object Header Address
    *(uint32_t *)(superblock + 68) = 0; // Cache Type
    *(uint32_t *)(superblock + 72) = 0; // Reserved
    *(uint64_t *)(superblock + 76) = btree_offset; // B-tree Address
    *(uint64_t *)(superblock + 84) = name_heap_offset; // Name Heap Address
    *(uint64_t *)(superblock + 44) = data_offset + compressed_offset; // End of File

    // B-tree (minimal, one node, 48 bytes)
    uint8_t btree[48] = {
        'T', 'R', 'E', 'E', // Signature
        0,    // Node Type (0 = group)
        1,    // Node Level (leaf)
        1, 0, // Entries Used
        0, 0, 0, 0, 0, 0, 0, 0, // Left Sibling (none)
        0, 0, 0, 0, 0, 0, 0, 0, // Right Sibling (none)
        0, 0, 0, 0, 0, 0, 0, 0, // Key 0 (offset in heap)
        0, 0, 0, 0, 0, 0, 0, 0  // Child Pointer (dataset offset)
    };
    *(uint64_t *)(btree + 32) = 0; // Key (offset in heap for "signal_data")
    *(uint64_t *)(btree + 40) = dataset_offset; // Child Pointer

    // Name Heap (64 bytes, minimal)
    uint8_t name_heap[64] = {
        'H', 'E', 'A', 'P', // Signature
        0, 0, 0, 0, // Version, reserved
        0, 0, 0, 0, 0, 0, 0, 0, // Data Segment Size
        0, 0, 0, 0, 0, 0, 0, 0, // Free List Head
        0, 0, 0, 0, 0, 0, 0, 0, // Data Segment Address
        's', 'i', 'g', 'n', 'a', 'l', '_', 'd', 'a', 't', 'a', 0 // Name
    };
    *(uint64_t *)(name_heap + 8) = 32; // Data Segment Size
    *(uint64_t *)(name_heap + 24) = name_heap_offset + 32; // Data Segment Address

    // Dataset Object Header (88 bytes)
    uint8_t dataset_header[88] = {
        1, 0,    // Version, reserved
        4, 0,    // Number of messages
        1, 0, 0, 0, // Reference count
        64, 0, 0, 0, // Header size
        // Message 1: Dataspace
        1, 0,    // Type: Dataspace
        16, 0,   // Size
        0,       // Flags
        2,       // Rank
        0, 0,    // Reserved
        0, 0, 0, 0, 0, 0, 0, 0, // Dim 1 (N)
        0, 0, 0, 0, 0, 0, 0, 0, // Dim 2 (M)
        // Message 2: Datatype
        0, 0,    // Type: Datatype
        8, 0,    // Size
        0,       // Class (fixed-point)
        4, 0, 0, // Size (float32)
        // Message 3: Data Layout
        3, 0,    // Type: Data Layout
        16, 0,   // Size
        0,       // Flags
        2,       // Version
        3,       // Layout class (chunked)
        0, 0, 0, 0, 0, 0, 0, 0, // Address
        0, 0, 0, 0, 0, 0, 0, 0, // Size
        // Message 4: Filter
        11, 0,   // Type: Filter
        8, 0,    // Size
        1, 0,    // Filter ID (DEFLATE)
        0, 0     // No client data
    };
    *(uint64_t *)(dataset_header + 20) = dim_n;
    *(uint64_t *)(dataset_header + 28) = dim_m;
    *(uint64_t *)(dataset_header + 56) = data_offset;
    *(uint64_t *)(dataset_header + 64) = compressed_offset;

    // Write all structures
    write(fd, superblock, 96);
    write(fd, btree, 48);
    write(fd, name_heap, 64);
    write(fd, dataset_header, 88);
    if (dsa_write(dsa_job, compressed_data, fd, data_offset, compressed_offset) != 0) {
        fprintf(stderr, "Error writing compressed data\n");
    }

    // Cleanup
    free(compressed_data);
    cleanup_dsa(dsa_job);
    qzTeardownSession(&session);
    qzClose(&session);
    close(fd);
}

int main() {
    float *circular_buffer = aligned_alloc(ALIGNMENT, BUFFER_SIZE * sizeof(float));
    if (!circular_buffer) {
        perror("Error allocating circular buffer");
        return 1;
    }
    for (int i = 0; i < BUFFER_SIZE; i++) {
        circular_buffer[i] = (float)i;
    }

    int head = BUFFER_SIZE / 2;
    int size = 1000000000;
    int capacity = BUFFER_SIZE;

    write_with_qat_dsa(circular_buffer, head, size, capacity, "output.h5",
                       "signal_data", 0, 1000000000, 1, 1000.0);

    free(circular_buffer);
    return 0;
}
